Package: @PKGNAME@
Type: Package
Title: @PKGTITLE@
Description: @PKGDESCRIPTION@
Version: @PKGVERSION@
Date: @DATE@
Author: @AUTHOR@
Maintainer: @MAINTAINER@
Depends: @GENOMENAME@
License: @LIC@
LazyLoad: yes
organism: @ORGANISM@
provider: @PROVIDER@
provider_version: @PROVIDERVERSION@
release_date: @RELEASEDATE@
release_name: @RELEASENAME@
source: @SRCDATAFILES@
biocViews: @ORGANISMBIOCVIEW@
