Package: QuasR
Type: Package
Title: Quantify and Annotate Short Reads in R
Version: 1.51.0
Date: 2025-06-23
Authors@R: c(person("Anita", "Lerch", role = c("aut")),
             person("Adam Alexander Thil", "SMITH", role = c("aut"), 
                    email = "alex.smith@fmi.ch", 
                    comment = c(ORCID = "0000-0003-4593-3042")),
             person("Charlotte", "Soneson", role = c("aut"), 
                    email = "charlottesoneson@gmail.com", 
                    comment = c(ORCID = "0000-0003-3833-2169")),
             person("Dimos", "Gaidatzis", role = c("aut")),
             person("Michael", "Stadler", role = c("aut", "cre"), 
                    email = "michael.stadler@fmi.ch", 
                    comment = c(ORCID = "0000-0002-2269-4934")))
Depends: R (>= 4.4), parallel, GenomicRanges, Rbowtie
Imports: methods, grDevices, graphics, utils, stats, tools,
        BiocGenerics, S4Vectors, IRanges, Biobase, Biostrings,
        BSgenome, Rsamtools (>= 2.13.1), GenomicFeatures, txdbmaker,
        ShortRead, BiocParallel, Seqinfo, rtracklayer, GenomicFiles,
        AnnotationDbi
Suggests: Gviz, BiocStyle, GenomeInfoDbData, GenomicAlignments,
        Rhisat2, knitr, rmarkdown, covr, testthat
LinkingTo: Rhtslib (>= 1.99.1)
Encoding: UTF-8
SystemRequirements: GNU make
Description: This package provides a framework for the quantification
    and analysis of Short Reads. It covers a complete workflow
    starting from raw sequence reads, over creation of alignments
    and quality control plots, to the quantification of genomic
    regions of interest. Read alignments are either generated through
    Rbowtie (data from DNA/ChIP/ATAC/Bis-seq experiments) or
    Rhisat2 (data from RNA-seq experiments that require spliced alignments),
    or can be provided in the form of bam files.
URL: https://bioconductor.org/packages/QuasR
License: GPL-2
biocViews: Genetics, Preprocessing, Sequencing, ChIPSeq, RNASeq,
        MethylSeq, Coverage, Alignment, QualityControl, ImmunoOncology
Archs: x64
VignetteBuilder: knitr
RoxygenNote: 7.3.2
BugReports: https://github.com/fmicompbio/QuasR/issues
git_url: https://git.bioconductor.org/packages/QuasR
git_branch: devel
git_last_commit: 331f3b2
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-11-04
NeedsCompilation: yes
Packaged: 2025-11-05 00:25:56 UTC; biocbuild
Author: Anita Lerch [aut],
  Adam Alexander Thil SMITH [aut] (ORCID:
    <https://orcid.org/0000-0003-4593-3042>),
  Charlotte Soneson [aut] (ORCID:
    <https://orcid.org/0000-0003-3833-2169>),
  Dimos Gaidatzis [aut],
  Michael Stadler [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-2269-4934>)
Maintainer: Michael Stadler <michael.stadler@fmi.ch>
