#' Quantify and Annotate Short Reads in R
#'
#' This package provides a pipeline for the quantification and analysis of 
#' short reads.
#'
#' See packageDescription('QuasR') for package details.
#' 
#' @author Anita Lerch, Charlotte Soneson, Dimos Gaidatzis and Michael Stadler
#' 
#' @aliases QuasR QuasR-package
#' @name QuasR-package
#' 
#' @keywords package
#' 
#' @seealso \code{\link{qAlign}}, \code{\link{qCount}}, \code{\link{qProfile}}, 
#'   \code{\link{qMeth}}, \code{\link{qQCReport}}
#' 
#' @useDynLib QuasR, .registration=TRUE
#' 
#' @examples 
#' \dontrun{
#' # see qCount, qMeth and qProfile manual pages for examples
#' example(qCount)
#' example(qMeth)
#' example(qProfile)
#' }
#' @keywords internal 
"_PACKAGE"
